/*
 * This file is part of LibQtTracker project
 *
 * Copyright (C) 2009, Nokia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
/*
 * tracker.cpp
 *
 *  Created on: Jan 5, 2009
 *      Author: Iridian Kiiskinen <ext-iridian.kiiskinen at nokia.com>
 */

#include "tracker_p.h"
#include "rdftrackerselect_p.h"
#include "rdftrackerupdate_p.h"
#include "../local/local_p.h"
#include "../../include/QtTracker/Tracker"
#include "../../util_p.h"

#include <uuid.h>


Q_DECLARE_METATYPE(QEventLoop::ProcessEventsFlags)

namespace SopranoLive
{
	namespace BackEnds
	{
		namespace Tracker
		{
			QString service_g;
			QString base_path_g;

			QString search_interface_g;
			QString search_path_g;

			QString resources_interface_g;
			QString resources_path_g;

			QString resources_class_path_g;
			QString resources_class_interface_g;

			QMutex tracker_global_lock_g;

			QStringList initReservedSparqlWords()
			{
				QStringList ret;
				ret << "id" << "uri";
				return ret;
			}
			QStringList reserved_sparql_words = initReservedSparqlWords();

			using namespace SopranoLive;


			void dispatchBatchSparqlUpdate(RDFUpdate const &update)
			{
				service()->executeQuery(update, BatchMode);
			}

			void dispatchBatchCommit()
			{
				TrackerService::get()->resources.BatchCommit();
			}

			UpdateResultOperations sparqlUpdateBlank(const QString& query)
			{
				QSharedPointer<TrackerService> tr = TrackerService::get();
				QDBusPendingReply<UpdateResultOperations> reply
						= tr->resources.SparqlUpdateBlank(query);
				tr->waitForFinished(reply);
				return reply.value();
			}

			void sync(bool block)
			{
				QSharedPointer<TrackerService> tr = TrackerService::get();
				int sync_id = tr->sync_id_counter++;
				if(tr->verbosity() > 1)
					tr->debug(2) << "syncing tracker (sync id =" << sync_id << ")";
				PERF_EVENT("before_sync %d", sync_id);
				QDBusPendingReply<> rep(tr->resources.Sync());
				if(block)
				{
					PERF_EVENT("before_waiting_sync %d", sync_id);
					tr->waitForFinished(rep);
					PERF_EVENT("after_waiting_sync %d", sync_id);
					if(tr->verbosity() > 2)
						tr->debug(2) << "done syncing tracker (sync id =" << sync_id << ")";
				}
			}

			void setTrackerVerbosity(int level)
			{
				TrackerService::get()->setVerbosity((unsigned)level);
			}

			bool initializeTrackerEnvironment()
			{
				registerQTrackerDbusTypes();

				service_g = "org.freedesktop.Tracker1";
				base_path_g = "/org/freedesktop/Tracker1";

				if(!trackerBus().interface()->isServiceRegistered(service_g)
					&& (trackerBus().interface()->startService(service_g)
						, !trackerBus().interface()->isServiceRegistered(service_g)))
					qCritical() << "cannot connect to org.freedesktop.Tracker1 service";

				search_interface_g = service_g + ".Search";
				search_path_g = base_path_g + "/Search";

				resources_interface_g = service_g + ".Resources";
				resources_path_g = base_path_g + "/Resources";

				resources_class_interface_g = service_g + ".Resources.Class";
				resources_class_path_g = base_path_g + "/Resources/Classes";
				return true;
			}

			class ClosingDBusConnection : public QDBusConnection
			{
			public:
				ClosingDBusConnection(const QDBusConnection& c)
					: QDBusConnection(c)
					{
					}
				~ClosingDBusConnection()
					{
						disconnectFromBus(name());
					}
			};

			QThreadStorage<ClosingDBusConnection*> dbus_tls;
			QAtomicInt counter = 0;

			QDBusConnection trackerBus(bool is_thread_private)
			{
				if(!is_thread_private)
					return QDBusConnection::sessionBus();

				// Create a separate D-Bus connection for each thread. Use ClosingDBusConnection so
				// that the bus gets disconnected when the thread storage is deleted.
				if (!dbus_tls.hasLocalData()) {
					QString id = QString::number(counter.fetchAndAddRelaxed(1))
							.prepend("libqttracker-dbus-");
					dbus_tls.setLocalData(new ClosingDBusConnection
							(QDBusConnection::connectToBus(QDBusConnection::SessionBus, id)));
				}
				return *dbus_tls.localData();
			}

			QThreadStorage<QWeakPointer<RDFGraphCache> *> tracker_tls;
			QThreadStorage<TLSDestroyer<QSharedPointer<RDFGraphCache> > *> tracker_persistent_tls;

			QSharedPointer<RDFGraphCache> TrackerService::getRDFGraphCache(bool persistent)
			{
				QSharedPointer<RDFGraphCache> ret;
				if(!tracker_tls.hasLocalData()
					// check if the weak pointer in local data is valid and store it to ret
					|| !(ret = *tracker_tls.localData()))
				{
					static bool initialized = false;
					if(!initialized)
					{
						QMutexLocker lock(&tracker_global_lock_g);
						if(!initialized)
							initialized = initializeTrackerEnvironment();
					}

					QSharedPointer<TrackerService> tracker
							= QSharedCreator<TrackerService>::create();
					tracker->setParent( tracker.toWeakRef() );
					ret = QSharedCreator<RDFGraphCache>::create(tracker);
					tracker_tls.setLocalData(new QWeakPointer<RDFGraphCache>(ret));
				}
				if(persistent && !tracker_persistent_tls.hasLocalData())
					tracker_persistent_tls.setLocalData
							(new TLSDestroyer<QSharedPointer<RDFGraphCache> >(ret));
				return ret;
			}

			QSharedPointer<TrackerService> TrackerService::get()
			{
				return getRDFGraphCache(false)->backend().sharedFromThis<TrackerService>();
			}

			RDFServicePtr tracker()
			{
				return TrackerService::getRDFGraphCache();
			}

			RDFServicePtr service()
			{
				return TrackerService::getRDFGraphCache(false);
			}

			RDFServicePtr trackerRelease()
			{
				if(!tracker_persistent_tls.hasLocalData())
					return RDFServicePtr();
				ClassUpdateDispatcher::destroyAll();
				RDFServicePtr ret = *tracker_persistent_tls.localData();
				tracker_persistent_tls.setLocalData(0);
				return ret;
			}

			void enableProcessEventsOnBlock
					( RDFServicePtr const &tracker_service, QEventLoop::ProcessEventsFlags flags)
			{
				tracker_service->setServiceAttribute("process_events_on_block"
						, QVariant::fromValue(flags));
			}

			void disableProcessEventsOnBlock(RDFServicePtr const &tracker_service)
			{
				tracker_service->setServiceAttribute("process_events_on_block", QVariant());
			}

			QString stringFromNode(Node const &node)
			{
				if(node.isLiteral())
					switch(node.literal().type())
					{
					case QVariant::Date:
						return node.literal().toDate().toString(Qt::ISODate);
					case QVariant::DateTime:
						return node.literal().toDateTime().toString(Qt::ISODate);
					case QVariant::Time:
						return node.literal().toTime().toString(Qt::ISODate);
					default: break;
					}
				return node.toString();
			}

			Node nodeFromString(QString const &value, CppType type = CppType())
			{
				unsigned id = type.metaTypeId();

				if(!value.size())
				{
					if(id == QVariant::String)
						return Node(LiteralValue(value));
					return Node();
				}

				// XXX: maybe fuller support of user types, now assume all user types are
				// resources
				if(id >= QVariant::UserType)
					return Node(QUrl::fromEncoded(value.toAscii()));

				switch(id)
				{
				case CppType::None: break;
				case QVariant::Bool: return Node(LiteralValue(value == "true"));
				case QVariant::Int: return Node(LiteralValue(value.toInt()));
				case QVariant::UInt: return Node(LiteralValue(value.toUInt()));
				case QVariant::LongLong: return Node(LiteralValue(value.toLongLong()));
				case QVariant::ULongLong: return Node(LiteralValue(value.toULongLong()));
				case QVariant::Double: return Node(LiteralValue(value.toDouble()));
				case QVariant::Char: break; // return Node(LiteralValue(QChar));
				case QVariant::Map: break; // return Node(LiteralValue(QVariantMap));
				case QVariant::List: break; // return Node(LiteralValue(QVariantList));
				case QVariant::String: break; // return Node(LiteralValue(QString));
				case QVariant::StringList: break; // return Node(LiteralValue(QStringList));

				case QVariant::ByteArray:
					return Node(LiteralValue(value.toAscii())); // TODO: really?

				case QVariant::BitArray: break; // return Node(LiteralValue(QBitArray));

				case QVariant::Date:
					return Node(LiteralValue(QDate::fromString(value, Qt::ISODate)));

				case QVariant::Time:
					return Node(LiteralValue(QTime::fromString(value, Qt::ISODate)));

				case QVariant::DateTime:
					return Node(LiteralValue(QDateTime::fromString(value, Qt::ISODate)));
				// As of 3.9.2009, tracker is permissive regarding IRI's, so are we.
				case QVariant::Url: return Node(QUrl::fromEncoded(value.toAscii()));
				//	case QVariant::Url:
				//		return Node(QUrl::fromEncoded(value.toAscii(), QUrl::StrictMode));
				default: break;
				}
				return Node(LiteralValue(value));
			}

			TrackerResource::TrackerResource
					( QSharedPointer<RDFServiceImplBase> const &service, Node const &node)
				: ResourceOfService(service, node), has_signaling_type()
			{
				graph()->debug(4) << "TrackerResource TR(" << this << ") created:"
						<< node.toString();
			}

			TrackerResource::~TrackerResource()
			{
				graph()->debug(4) << "TrackerResource TR(" << this << ") destroyed:"
						<< node().toString();
			}

			TrackerService *TrackerResource::tracker() const
			{
				return static_cast<TrackerService *>(graph_.data());
			}

			void TrackerResource::removeInterface(QUrl const &interface_name) const
			{
				// tracker properly manages removal of all dependent triples
				graph_->deleteTriple(node(), rdf::type::iri(), interface_name);
			}

			LiveNodeModelBasePtr TrackerResource::implGetProperties(RDFDerivedSelect *columns
					, TripleModulePtr const &triple_module) const
			{
				RDFProperty &property_data = (*columns)[0].refPropertyData();

				bool implementing_live = false;
				if((property_data.strategy() & RDFStrategy::Live) && (columns->size() == 1))
				{
					property_data.disableStrategyFlags(RDFStrategy::LiveMask);
					implementing_live = true;
				}

				LiveNodeModelBasePtr ret = ResourceOfService::implGetProperties
						(columns, triple_module);

				if(implementing_live)
				{
					TripleModelPtr trip_ret = ret.staticCast<TripleModel>();
					SharedTripleUpdateContext ctxt(new TripleUpdateContext
							(trip_ret, property_data.asStatementWith(sourceVariable())));
					if(tracker()->registerForTripleUpdates(ctxt))
						trip_ret->doEnableStrategyFlags(RDFStrategy::Live);
					else
						graph()->warning() << "live property models with non-explicit predicates"
								" not supported, non-live model returned. Specify an explicit "
								"predicate.";
				}
				return ret;
			}

/*
			QSharedPointer<ContentModule> TCM_Model_Property::copy() const
			{
				return QSharedCreator<TCM_Model_Property>::create(*this);
			}
*/

			// if query_id is supplied, the query is flushed and executed
			// when attached. Otherwise, query has to be explicitly flushed.
			TrackerContentModule::TrackerContentModule
					( QSharedPointer<TrackerService> service, RDFDerivedSelect const &columns
					, bool flush_on_attach, ContentModule::ChangePolicy change_policy)
				: ContentModule_Simple(change_policy)
				, service_(service)
				, columns_(columns)
				, query_id_(0)
				, flush_on_attach_(flush_on_attach)
				, completion_watcher_(), data_watcher_()
			{
				//service_->debug(4) << "TrackerContentModule TMS(" << this << ") created";
			}
			TrackerContentModule::TrackerContentModule(TrackerContentModule const &cp)
				: ContentModule_Simple(cp)
				, service_(cp.service_)
				, columns_(cp.columns_)
				, query_id_()
				, flush_on_attach_(cp.flush_on_attach_)
				, completion_watcher_(), data_watcher_()
			{
				//service_->debug(4) << "TrackerContentModule TMS(" << this << ") copied"
				//		" from TMS(" << &cp << ")";
			}
			TrackerContentModule::~TrackerContentModule()
			{
				//service_->debug(4) << "TrackerContentModule TMS(" << this << ") destroyed";
			}


			QString TrackerContentModule::getQueryText() const
			{
				return model()->columns_.getQuery(reserved_sparql_words);
			}

			void TrackerContentModule::attached(RowStoreModel &rowstoremodel)
			{
				ContentModule_Simple::attached(rowstoremodel);
				model()->setColumnInformation(columns_);
				columns_ = RDFDerivedSelect();
				if(flush_on_attach_)
					model()->refreshModel(LiveNodeModel::Flush);
			}
			void TrackerContentModule::detached(RowStoreModel &rowstoremodel)
			{
				columns_ = model()->columns_;
				ContentModule_Simple::detached(rowstoremodel);
				completion_watcher_.clear();
				data_watcher_.clear();
			}

			ModelCommons::ReadyFlags TrackerContentModule::refreshModel_chain
					(LiveNodeModel::RefreshMode mode, InterfaceChain<ModelCommons> chain)
			{
				if(!service_->resources.isValid())
					return service_->warning() << service_->resources.lastError(), ReadyFlags();

				ReadyFlags ret = ContentModule_Simple::refreshModel_chain(mode, chain);

				if(mode & LiveNodeModel::Flush)
				{
					setUpToDate(false);
					QString query = getQueryText();
					PERF_EVENT("before_query %d", service_->query_id_counter);
					QDBusPendingReply<QVector<QStringList> > reply
							= service_->resources.SparqlQuery(query);
					data_watcher_ = QSharedPointer<QDBusPendingCallWatcher>
						(new QDBusPendingCallWatcher(reply), &QObject::deleteLater);
					completion_watcher_ = data_watcher_;
					QObject::connect(completion_watcher_.data(), SIGNAL(finished(QDBusPendingCallWatcher*))
							 , model(), SLOT(refreshModelSignal()));
					query_id_ = service_->query_id_counter++;
					if(service_->verbosity() > 1)
						service_->debug(2) << "modeling query (id =" << query_id_ << "):\n"
								<< query << "\n";
					// prevent refreshModelImpl from signaling
				}
				if(!upToDate())
				{
					// no watcher_ means query was aborted
					if(data_watcher_
							&& (mode & LiveNodeModel::Block || completion_watcher_->isFinished()))
					{
						// store watcher as local, because it might get cleaned at detach
						QSharedPointer<QDBusPendingCallWatcher> data_watcher;
						qSwap(data_watcher, data_watcher_);

						// Disconnect the watcher so that waitForFinished won't call
						// refreshModel again.
						completion_watcher_->disconnect();
						PERF_EVENT("before_waiting %d", query_id_);
						service_->waitForFinished(*completion_watcher_);
						PERF_EVENT("after_waiting %d", query_id_);
						completion_watcher_.clear();
						updatePending();

						if(!data_watcher->isError())
						{
							QDBusPendingReply<QVector<QStringList> > reply = *data_watcher;
							PERF_EVENT("before_transitionrows %d", query_id_);
							model()->transitionRowsImpl
								(0, model()->row_store.size(), RowStore(reply.value()));
							PERF_EVENT("after_transitionrows %d", query_id_);
						} else
							model()->emitErrorImpl
									( data_watcher->error().message(), 0, QModelIndex()
									, !(mode & LiveNodeModel::Block));

						// note that we might have gotten detached in the meanwhile.
						// Dont assume anything after transitionRowsImpl finishes.

						setUpToDate(true);
						if(service_->verbosity() >= 1)
						{
							if(!data_watcher->isError())
							{
								if(service_->verbosity() >= 2)
									service_->debug(2)
											<< "done modeling query (id =" << query_id_ << "):";
								if(service_->verbosity() >= 4)
									service_->debug(4)
											<< "   query was:\n" + getQueryText() << "\n";
								if(service_->verbosity() >= 3)
									service_->debug(3) << "   results:\n" << *model();
							} else
							{
								service_->warning()
										<< "error while modeling query (id =" << query_id_
										<< "):\n" << data_watcher->error().message();
								if(service_->verbosity() == 1 || service_->verbosity() >= 4)
									service_->warning()
											<< "   query was:\n" << getQueryText() << "\n";
							}
						}
					} else
					{
						ret = NothingReady;
					}
				}
				return ret;
			}

			bool TrackerContentModule::alterStrategy_chain
					( RDFStrategyFlags flags, RDFStrategyFlags mask, QModelIndex const &index
					, InterfaceChain<ModelCommons> chain)
			{
				int svc_v = service_->verbosity() >= 5;
				if(svc_v >= 5) model()->debug(2, flags, mask)
						<< "TrackerContentModule::alterStrategy" << model()->strategy();

				// diagnostics. Handle Running.
				RDFStrategyFlags running_resp = mask & RDFStrategy::Running;

				// delegation
				if(!ContentModule_Simple::alterStrategy_chain
						(flags, mask & ~running_resp, index, chain))
					return false;

				// implementation
				if(running_resp)
				{
					if(!(flags & RDFStrategy::Running))
					{
						if(svc_v >= 5) model()->debug(2, flags, mask)
								<< "TrackerContentModule::alterStrategy: "
										"disabling Running, stopping query";
						data_watcher_.clear();
						completion_watcher_.clear();
						// FIXME: ModelReady should not be set when query is aborted, but this
						// solution will end up setting it when refreshModel is next time called.
						setUpToDate(true);
					} else
					if(!(model()->strategy_ & RDFStrategy::ModelReady))
					{
						if(svc_v >= 5) model()->debug(2, flags, mask)
								<< "TrackerContentModule::alterStrategy: "
										"enabling Running, executing query";
						model()->refreshModel();
					}
				}
				return true;
			}


			QSharedPointer<ContentModule> TrackerContentModule::copy() const
			{
				return QSharedCreator<TrackerContentModule>::create(*this);
			}

			TrackerUpdateModule::TrackerUpdateModule
					( QSharedPointer<TrackerService> service
					, RDFDerivedSelect const& columns
					, RDFUpdate const &update
					, bool flush_on_attach, bool batch_mode, bool sync_on_commit
					, ContentModule::ChangePolicy change_policy)
				: TrackerContentModule(service, columns, flush_on_attach, change_policy)
				, update_(update)
				, batch_mode_(batch_mode)
				, sync_on_commit_(sync_on_commit)
				, my_index_(0)
			{
			}

			TrackerUpdateModule::TrackerUpdateModule(TrackerUpdateModule const &cp)
				: QObject(), TrackerContentModule(cp)
				, update_(cp.update_)
				, batch_mode_(cp.batch_mode_), sync_on_commit_(cp.sync_on_commit_)
				, my_index_(0)
			{
			}

			TrackerUpdateModule::~TrackerUpdateModule()
			{
				// Remove this module from the sister lists of the sister modules
				foreach(TrackerUpdateModule *sister, sister_modules_)
				{
					if (sister != 0)
						sister->sister_modules_[my_index_] = 0;
				}
			}

			QString TrackerUpdateModule::getQueryText() const
			{
				return update_.getQuery(reserved_sparql_words);
			}

			ModelCommons::ReadyFlags TrackerUpdateModule::refreshModel_chain
					( LiveNodeModel::RefreshMode mode, InterfaceChain<ModelCommons> chain)
			{
				if(!service_->resources.isValid())
					return service_->warning() << service_->resources.lastError(), ReadyFlags();

				ReadyFlags ret = ContentModule_Simple::refreshModel_chain(mode, chain);

				if(mode & LiveNodeModel::Flush)
				{
					setUpToDate(false);
					constructPendingCall();
				}
				if(!upToDate())
				{
					if(mode & LiveNodeModel::Block ||
							(completion_watcher_ && completion_watcher_->isFinished()))
					{
						// store watcher as local, because it might get cleaned at detach
						QSharedPointer<QDBusPendingCallWatcher> data_watcher;
						// This is ok even if the sister models still need the pending call,
						// since they hold QSharedPointers to it.
						qSwap(data_watcher, data_watcher_);

						if(sync_on_commit_)
							// completion_watcher_ will be assigned to us in someone's runSync
							service_->waitForFinished(*data_watcher);

						// Disconnect the watcher from us so that waitForFinished won't call
						// refreshModel again. (However, keep the connection between the
						// watcher and the sister modules.)
						completion_watcher_->disconnect(model());
						service_->waitForFinished(*completion_watcher_);
						completion_watcher_.clear();


						if(data_watcher->isError())
						{
							model()->emitErrorImpl
									( data_watcher->error().message(), 0, QModelIndex()
									, !(mode & LiveNodeModel::Block));
						}
						else if (!batch_mode_)
						{
							// If the update contained blank nodes, the result contains the
							// generated ID's for those.

							// D-Bus signature of the result: aaa{ss}
							// a{ss} is a map: blank node id -> generated uri
							// middle "a" is because an insert block may have multiple solutions
							// highest-level "a" is because the query may have multiple insert
							// statements

							UpdateResultOperations const result = QDBusPendingReply
									<UpdateResultOperations>(*data_watcher).value();
							if (result.size() > my_index_)
							{
								updatePending();
								// This TrackerUpdatemodule needs to handle the insert statement
								// with the index my_index_.
								UpdateResultSolutions const generated_blanks = result[my_index_];
								if (generated_blanks.size() > 0)
								{
									QStringList keys;
									// We assume that all the maps have the same keys
									// = blank node names. They will be the columns of the model.
									foreach(RDFDerivedColumn const &dc, model()->columns_.columns())
										keys << dc.name();

									QVector<QStringList> data_for_model(generated_blanks.size());

									for (int i=0; i<generated_blanks.size(); ++i)
									{
										UpdateResultBlanksMap const solution = generated_blanks[i];
										foreach(const QString& key, keys)
										{
											data_for_model[i] << solution[key];
										}
									}

									model()->transitionRowsImpl
										(0, model()->row_store.size(), RowStore(data_for_model));
								}
							}
						}

						setUpToDate(true);
						if(service_->verbosity() >= 1)
						{
							if(!data_watcher->isError())
							{
								if(service_->verbosity() >= 2)
									service_->debug(2)
											<< "done modeling update query (update id =" << query_id_ << "):";
								if(service_->verbosity() >= 4)
									service_->debug(4)
											<< "   query was:\n" + getQueryText() << "\n";
								if(service_->verbosity() >= 3)
									service_->debug(3) << "   results:\n" << *model();
							} else
							{
								service_->warning()
										<< "error while modeling update query (update id =" << query_id_
										<< "):\n" << data_watcher->error().message();
								if(service_->verbosity() == 1 || service_->verbosity() >= 4)
									service_->warning()
											<< "   query was:\n" << getQueryText() << "\n";
							}
						}
					} else
					{
						ret = NothingReady;
					}
				}
				return ret;
			}

			void TrackerUpdateModule::setSisterModules
					( QList<TrackerUpdateModule*> const &sister_modules, int ix)
			{
				sister_modules_ = sister_modules;
				my_index_ = ix;
			}

			void TrackerUpdateModule::runSync()
			{
				service_->sync_id_counter++;
				completion_watcher_ = QSharedPointer<QDBusPendingCallWatcher>
								(new QDBusPendingCallWatcher(service_->resources.Sync()));
				foreach(TrackerUpdateModule *sister, sister_modules_)
				{
					if (sister != 0)
					{
						sister->completion_watcher_ = completion_watcher_;
						QObject::connect
								( completion_watcher_.data()
								, SIGNAL(finished(QDBusPendingCallWatcher*))
								, sister->model(), SLOT(refreshModelSignal()));
					}
				}
			}

			// If none of the sister models has initiated an asynchronous call over D-Bus, do it
			// now. If somebody has, just start watching it.
			void TrackerUpdateModule::constructPendingCall()
			{
				if (completion_watcher_.data() == 0)
				{
					// This is the first sister module to initiate the creation of the
					// pending call.
					QString query = getQueryText();

					QDBusPendingReply<> reply;
					if (batch_mode_)
					{
						// First call the SparqlBatchUpdate, then BatchCommit. The update is
						// finished when BatchCommit returns.
						service_->resources.BatchSparqlUpdate(query);
						reply = service_->resources.BatchCommit();
					} else
					{
						// Normal mode; just call SparqlUpdate. The update is finished when it
						// returns.
						reply = service_->resources.SparqlUpdateBlank(query);
					}
					data_watcher_ = QSharedPointer<QDBusPendingCallWatcher>
							(new QDBusPendingCallWatcher(reply));

					if(!sync_on_commit_)
						completion_watcher_ = data_watcher_;
					else
						QObject::connect(data_watcher_.data(), SIGNAL(finished(QDBusPendingCallWatcher*))
										 , this, SLOT(runSync()));

					// Give the same watcher to the sister models as a shared pointer
					foreach(TrackerUpdateModule *sister, sister_modules_)
					{
						if (sister != 0 && sister != this)
						{
							sister->data_watcher_
								= QSharedPointer<QDBusPendingCallWatcher>(data_watcher_);
							sister->completion_watcher_
								= QSharedPointer<QDBusPendingCallWatcher>(completion_watcher_);
						}
					}
					query_id_ = service_->update_id_counter++;
					if(service_->verbosity() > 1)
						service_->debug(2)
								<< (sync_on_commit_
									? "modeling and syncing"
									: "modeling")
								<< "update query (update id =" << query_id_ << "):\n"
								<< query << "\n";

				}

				// Connect the pending call watcher in any case (whether we were the creator
				// of the pending call or not.
				if(completion_watcher_.data())
					QObject::connect
							( completion_watcher_.data()
							, SIGNAL(finished(QDBusPendingCallWatcher*))
							, model(), SLOT(refreshModelSignal()));
			}

			TCM_TextQuery::TCM_TextQuery
					( QSharedPointer<TrackerService> service, QString const &query_text
					, RDFDerivedSelect const &columns, bool flush_on_attach)
				: TrackerContentModule(service, columns, flush_on_attach)
				, query_text_(query_text)
			{}

			QSharedPointer<TCM_TextQuery> TCM_TextQuery::create
					( QSharedPointer<TrackerService> service, QString const &query_text
					, QStringList const &columns, bool flush_on_attach)
			{
				RDFDerivedSelect sel;
				for(QStringList::const_iterator fi = columns.begin(); fi != columns.end(); ++fi)
					sel.addColumn(RDFVariable(*fi));
				return QSharedCreator<TCM_TextQuery>::create
						(service, query_text, sel, flush_on_attach);
			}

			QSharedPointer<ContentModule> TCM_TextQuery::copy() const
			{
				return QSharedCreator<TCM_TextQuery>::create(*this);
			}

			QString TCM_TextQuery::getQueryText() const
			{
				return query_text_;
			}


			TrackerService::TrackerService()
				: RDFServiceBasic("SopranoLive; Tracker backend", Contexts::QtTracker::iriList())
				, resources(service_g, resources_path_g, trackerBus(), this)
				, id_counter(0), query_id_counter(0), update_id_counter(0), sync_id_counter(0)
				, local_service_(Local::LocalService::getRDFGraphCache(false, true))
			{
				setVerbosity(1);

				stem = Contexts::QtTracker::encodedIri();

				service_context_data_->select_prototype_ = RDFTrackerSelectPrivate::prototype();
				service_context_data_->update_prototype_ = RDFTrackerUpdatePrivate::prototype();

				setDefaultTimeout(-1);

				loadTrackerPrefixes();

				// << "TrackerService TS(" << this << ") created";
			}

			TrackerService::~TrackerService()
			{
				// << "TrackerService TS(" << this << ") destroyed";
			}

			QUrl TrackerService::serviceIri() const
			{
				return QUrl("http://www.tracker-project.org/");
			}

			bool TrackerService::waitForFinishedPendingCall
					( QDBusPendingCall &pending_call
					, QEventLoop::ProcessEventsFlags flags) const
			{
				QDBusPendingCallWatcher watcher(pending_call);
				QEventLoop loop;
				QObject::connect(&watcher, SIGNAL(finished(QDBusPendingCallWatcher *))
						, &loop, SLOT(quit()));
				loop.exec(flags);
				return true;
			}

			ClassUpdateDispatcher *TrackerService::dispatcherFromClass(QString const &class_uri)
			{
				QString2Dispatcher::iterator i = class2dispatcher.find(class_uri);
				if(i == class2dispatcher.end())
				{
					QPair<QString, QString> onto_class = splitToPrefixSuffix(class_uri);
					ClassUpdateDispatcher *disp;
					if(onto_class.first.size())
						disp = ClassUpdateDispatcher::create
							(class_uri, onto_class.first, onto_class.second);
					else
						warning() << "while registering class signals: "
								"no ontology prefix registered for" << class_uri;
					i = class2dispatcher.insert(class_uri, disp);
				}
				return *i;
			}
			ClassUpdateDispatcher *TrackerService::dispatcherFromProperty(QString const &property)
			{
				QString2Dispatcher::iterator i = property2dispatcher.find(property);
				if(i == property2dispatcher.end())
				{
					QList<LiveNode> domainresult
							= tracker()->modelVariable
								(RDFVariable(QUrl(property)).property<rdfs::domain>());
					QString domain = domainresult.size() ? domainresult[0].toString() : QString();
					ClassUpdateDispatcher *disp;
					if(domain.size())
						disp = dispatcherFromClass(domain);
					else
						warning() << "QtTracker: while registering class signals: "
								"predicate" << property << "has no domain";
					i = property2dispatcher.insert(property, disp);
				}
				return *i;
			}

			QVariant TrackerService::serviceAttribute(QString const &attribute_name) const
			{
				if(attribute_name == "query_count")
					return QVariant(query_id_counter);
				else if(attribute_name == "update_count")
					return QVariant(update_id_counter);
				else if(attribute_name == "sync_count")
					return QVariant(sync_id_counter);
				return RDFServiceBasic::serviceAttribute(attribute_name);
			}

			QVariantMap TrackerService::serviceAttributes() const
			{
				QVariantMap ret = RDFServiceBasic::serviceAttributes();
				ret["query_count"] = QVariant(query_id_counter);
				ret["update_count"] = QVariant(update_id_counter);
				ret["sync_count"] = QVariant(sync_id_counter);
				return ret;
			}

			int trackerDefaultTimeout()
			{
				return TrackerService::get()->defaultTimeout();
			}

			int TrackerService::defaultTimeout() const
			{
				return attributes_["tracker_timeout"].toInt();
			}

			void setTrackerDefaultTimeout(int timeout)
			{
				TrackerService::get()->setDefaultTimeout(timeout);
			}

			void TrackerService::setDefaultTimeout(int timeout)
			{
				attributes_["tracker_timeout"] = QVariant(timeout);
			}

			QStringList TrackerService::inferNotifyTypes(RDFVariable const &resource) const
			{
				QStringList ret;

				QVector<RDFProperty> deriveds = resource.derivedProperties();
				for(QVector<RDFProperty>::const_iterator di = deriveds.constBegin()
											, diend = deriveds.constEnd(); di != diend; ++di)
					if( (di->strategy() & RDFStrategy::Object)
							&& di->target().metaIsDefinite()
							&& di->target().metaStrategy() & Notify)
						// TODO: assumes all types are partial information types, add support
						// for distinguishing a fully defining type from partial types
						ret << di->target().metaValue().toString();
				return ret;
			}

			TripleUpdateContext::Mode TrackerService::registerForTripleUpdates
					(SharedTripleUpdateContext const &context)
			{
				// variables describing the desired triples. We're mostly
				// interested in the types of subj, which are stored in
				// subject's derived properties to get the proper class handler
				RDFVariable subj = context->triple_mask_.subject();
				RDFVariable pred = context->triple_mask_.predicate();
				RDFVariable obj = context->triple_mask_.object();

				QString pred_str, subj_str;
				QStringList notify_types;

				ClassUpdateDispatcher::Events events;

				if(pred.metaValue() == rdf::type::iri())
				{
					if(obj.metaStrategy() & Notify)
						notify_types << obj.metaValue().toString();
					events = ClassUpdateDispatcher::AddsAndRemoves;
				} else
				{
					if(subj.metaIsDefinite())
						subj_str = subj.metaValue().toString();
					if(pred.metaIsDefinite())
					{
						pred_str = pred.metaValue().toString();
						events = ClassUpdateDispatcher::Updates;
					} else
						events = ClassUpdateDispatcher::AllEvents;
				}

				// << "Registering signals for "
				//		"triples(" << subj_str << ", " << pred_str << ", -)";

				if(!notify_types.size()
						&& !(notify_types = inferNotifyTypes(subj)).size()
					// && ! make dynamic query for type
					// TODO: handle synchronization issues! The registration for updates must
					// happen before the query actual gets executed, otherwise we might miss
					// some changes
					)
				{
					// TODO: output error?
					// << "No signaler found";
					return TripleUpdateContext::Inactive;
				}

				// << "Signaler found from types(" << notify_type << ")";

				foreach(QString const &notify_type, notify_types)
				{
					dispatcherFromClass(notify_type)->registerForEvents
							(context, events, subj_str, pred_str);
				}

				return context->mode_;
			}

			void TrackerService::loadTrackerPrefixes()
			{
				loadPrefixes();
				// TODO: replace hard-coded prefixes with something read from a file, or
				//  included from a build-time autogenerated header

				const char *ontologies[][2] =
				{
					{ "nrl", "http://www.semanticdesktop.org/ontologies/2007/08/15/nrl#" }
					// nepomuk ontologies http://www.semanticdesktop.org/ontologies/
				,	{ "nrl", "http://www.semanticdesktop.org/ontologies/2007/08/15/nrl#" }
				,	{ "nao", "http://www.semanticdesktop.org/ontologies/2007/08/15/nao#" }
				,	{ "nie", "http://www.semanticdesktop.org/ontologies/2007/01/19/nie#" }
				,	{ "nfo", "http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#" }
				,	{ "nco", "http://www.semanticdesktop.org/ontologies/2007/03/22/nco#" }
				,	{ "nmo", "http://www.semanticdesktop.org/ontologies/2007/03/22/nmo#" }
				,	{ "ncal", "http://www.semanticdesktop.org/ontologies/2007/04/02/ncal#" }
				,	{ "nexif", "http://www.semanticdesktop.org/ontologies/2007/05/10/nexif#" }
				,	{ "nid3", "http://www.semanticdesktop.org/ontologies/2007/05/10/nid3#" }
				,	{ "pimo", "http://www.semanticdesktop.org/ontologies/2007/11/01/pimo#" }
				,	{ "tmo", "http://www.semanticdesktop.org/ontologies/2008/05/20/tmo#" }
				,	{ "nexif", "http://www.semanticdesktop.org/ontologies/2007/05/10/nexif#" }

					// tracker ontologies
				,	{ "nmm", "http://www.tracker-project.org/temp/nmm#" }
				,	{ "mfo", "http://www.tracker-project.org/temp/mfo#" }
				,	{ "mto", "http://www.tracker-project.org/temp/mto#" }
				,	{ "tracker", "http://www.tracker-project.org/ontologies/tracker#" }
				,	{ "fts", "http://www.tracker-project.org/ontologies/fts#" }
				,	{ "poi", "http://www.tracker-project.org/ontologies/poi#" }

					// maemo ontologies
				,	{ "maemo", "http://maemo.org/ontologies/tracker#" }

					// misc ontologies
				,	{ "exif", "http://www.kanzaki.com/ns/exif#" }
				,	{ "protege", "http://protege.stanford.edu/system#" }
				,	{ "xesam", "http://freedesktop.org/standards/xesam/1.0/core#" }
				};
				for(unsigned i = 0; i < sizeof(ontologies) / sizeof(const char*[2]); ++i)
					addPrefix(ontologies[i][0], QUrl(ontologies[i][1]));
			}

			// graphs not supported
			QVector<QStringList> TrackerService::rawExecuteQueryOnGraph
					( QString const &query, int type, RDFGraphImplBase const &/*graph*/)
			{
				if(!resources.isValid())
				{
					warning() << resources.lastError().message();
					return QVector<QStringList>();
				}

				if(type != RDFQuery::SPARQL
						&& type != RDFQuery::SPARQLUpdate && type != BatchSPARQLUpdate)
				{
					warning() << "unrecognized query type";
					return QVector<QStringList>();
				}

				if(verbosity() >= 2)
					debug(2) << "executing"  << (type == RDFQuery::SPARQL
								? "select"
								: (type == RDFQuery::SPARQLUpdate ? "update" : "batch update"))
							<< "query:" << query;

				QVector<QStringList> ret;

				QString reply_error = QString();
				if(type == RDFQuery::SPARQL)
				{
					QDBusPendingReply<QVector<QStringList> > reply = resources.SparqlQuery(query);

					waitForFinished(reply);

					if(reply.isError())
						reply_error = reply.error().message();
					ret = reply.value();
				} else
				if(type == RDFQuery::SPARQLUpdate || type == BatchSPARQLUpdate)
				{
					QDBusPendingReply<> reply =
						(type == RDFQuery::SPARQLUpdate
								? resources.SparqlUpdate(query)
								: resources.BatchSparqlUpdate(query));
					if(verbosity() == 2)
						debug(2) << "   async op, ignoring any errors "
								"(use verbose >= 3 to catch them)";
					if(verbosity() > 2)
					{
						waitForFinished(reply);
						if(reply.isError())
							reply_error = reply.error().message();
					}
					else
						return ret;
				}

				if(verbosity())
				{
					if(reply_error.size())
					{
						warning() << "error while executing"
								<< (type == RDFQuery::SPARQL ? "select" : "update")
								<< "query: " << reply_error;
						if(verbosity() == 1 || verbosity() >= 4)
							warning(1) << "   query was:\n" + query;
					}
					else if(verbosity() >= 2)
						debug(2) << "   query results:" << ret;
				}
				return ret;
			}

			void TrackerService::rawInsertTripleToGraph
					( QString const &subject, QString const &predicate, QString const &object
					, RDFGraphImplBase const &/*graph*/)
			{
				critical() << "deprecated API: not implemented";
				return;
			}

			void TrackerService::rawDeleteTripleFromGraph
					( QString const &subject, QString const &predicate, QString const &object
					, RDFGraphImplBase const &/*graph*/)
			{
				critical() << "deprecated API: not implemented";
				return;
			}

			void TrackerService::rawLoadOnGraph
					( QUrl const &file_uri, RDFGraphImplBase const &/*graph*/)
			{
				if(!resources.isValid())
				{
					warning() << resources.lastError().message();
					return;
				}

				QDBusPendingReply<> reply = resources.Load(file_uri.toString());
				if(verbosity() != 0)
				{
					if(verbosity() >= 2)
						debug() << "Loading uri: " << file_uri;
					waitForFinished(reply);
					if(reply.isError())
						warning(1) << "while executing load(" << file_uri << "), tracker error: "
								<< reply.error().message();
				}
			}

			QUrl TrackerService::createUniqueIri(QString stem_text)
			{
				uuid_t uuid;
				uuid_generate(uuid);
				char uuid_string[37];
				uuid_unparse(uuid, uuid_string);
				return stem + stem_text + "/" + QString(uuid_string);
			}

			TrackerResource *TrackerService::newLiveResource(Node const &node)
			{
				return new TrackerResource(sharedFromThis<RDFGraphBasic>(), node);
			}

			Node TrackerService::rawNodeFromString(QString string, CppType type)
			{
				return nodeFromString(string, type);
			}

			QString TrackerService::rawStringFromNode(Node const &node)
			{
				return stringFromNode(node);
			}

			RDFStrategyFlags TrackerService::preprocessQuery(RDFStrategyFlags strategy)
			{
				// nullify all requested features with Allow specification
				return strategy & ~RDFStrategy::AllowsMask
						& ~(strategy >> (RDFStrategy::offset_AllowStreaming
								- RDFStrategy::offset_Streaming));
			}

			LiveNodes TrackerService::modelQuery
					( QString const &select, QStringList const &columns
					, RDFStrategyFlags strategy, QSharedPointer<QObject> const &parent
					, TripleModulePtr const &triple_module)
			{
				strategy = preprocessQuery(strategy);

				TripleModelPtr ret = newTripleModel
					// tracker backend doesn't implement any strategies
					( 0
					, TCM_TextQuery::create
						( sharedFromThis<TrackerService>()
						, select, columns
						, false)
					, triple_module
					, parent
					);

				ret->alterStrategy(strategy, RDFStrategy::Running);

				return LiveNodes(ret);
			}

			LiveNodeModelBasePtr TrackerService::baseImplModelQuery
					( RDFDerivedSelect *select, RDFStrategyFlags strategy
					, QSharedPointer<QObject> const &parent
					, TripleModulePtr const &triple_module)
			{
				strategy = preprocessQuery(strategy);

				TripleModelPtr ret = newTripleModel
					( 0
					, QSharedCreator<TrackerContentModule>::create
						( sharedFromThis<TrackerService>()
						, RDFDerivedSelect().swap(*select)
						, false
						)
					, triple_module
					, parent
					);

				ret->alterStrategy(strategy, RDFStrategy::Running);

				return ret;
			}

			QList<LiveNodes> TrackerService::executeQuery
					( RDFQuery const &update, RDFStrategyFlags strategy
					, QSharedPointer<QObject> const &parent)
			{
				Q_UNUSED(parent);

				// FIXME: MAJOR, this function and functionality underneath should go
				// considerable refactoring


				if(transaction_)
				{
					// We have an active transaction; don't execute the update
					// now but add it to the transaction.
					RDFUpdate const &real_update = update.toRDFUpdate();
					if(!real_update.isEmpty())
					{
						transaction_->update_.appendUpdate(real_update);
					} else
						warning(1) << "TrackerService::executeQuery: "
								"executing select queries not supported, returning null";
					// TODO: we should return something that can emit completion signals
					// even if we're in a tx as that's perfectly viable
					return QList<LiveNodes>();
				}

				if(strategy &
						(LiveNodeModel::Streaming | LiveNodeModel::Windowed | RDFStrategy::Live))
					warning(1) << "executeQuery: requested strategy (streaming, windowed or live)"
							" not supported, falling back to DefaultStrategy";

				RDFUpdate const &real_update = update.toRDFUpdate();
				// TODO: only support sparql update, fix.
				// TODO: also, handle the transaction more cleanly.
				if(real_update.isEmpty())
					warning(1) << "TrackerService::executeQuery: "
							"executing select queries not supported, returning null";

				int operation_count = real_update.operations().size();
				// if update has no operations, dont perform any queries
				if(!operation_count)
				{
					warning(2) << "TrackerService::executeQuery: null update, returning empty list";
					return QList<LiveNodes>();
				}

				RDFUpdate tracker_update = service_context_data_->update(real_update);

				if(strategy & RDFStrategy::DiscardUpdateResults)
				{
					rawExecuteQueryOnGraph
						( tracker_update.getQuery(reserved_sparql_words, RDFQuery::SPARQLUpdate)
						, (strategy & BatchMode)
							? int(BatchSPARQLUpdate)
							: int(RDFQuery::SPARQLUpdate)
						, *this);
					if(strategy & SyncOnCommit)
						// TODO: add support for direct blocking updates?
						sync(false);
					return QList<LiveNodes>();
				}

				// If we're doing batch update, no data will be returned as models.
				// Just create one model and a TrackerUpdateModule for it.
				if (strategy & BatchMode)
				{
					LiveNodes ret = LiveNodes(newTripleModel
							( strategy & 0
							, QSharedCreator<TrackerUpdateModule>::create
									( sharedFromThis<TrackerService>()
									, RDFDerivedSelect()
									, tracker_update
									, true
									, strategy & BatchMode
									, strategy & SyncOnCommit)
							, TripleModulePtr()
							, parent));
					return QList<LiveNodes>() << ret;
				}

				QList<RDFUpdate::Operation> ops = tracker_update.operations();

				// The modules cannot contain QSharedPointers to each other, otherwise they would
				// keep each other alive. Hence this complication.
				// TODO: replace with weak pointers?
				QList<QSharedPointer<TrackerUpdateModule> > update_modules;
				QList<TrackerUpdateModule*> update_modules_raw;
				for (int i = 0; i < operation_count; ++i)
				{
					// Check which are the blank variables in the operation; they will be the
					// columns of the returned model
					RDFDerivedSelect columns;
					QSet<QString> blank_variables;
					QList<RDFVariableStatement> inserts = ops[i].inserts();
					foreach(const RDFVariableStatement& stmt, inserts)
					{
						RDFVariable subject = stmt.subject();
						RDFVariable object = stmt.object();
						if (subject.metaIsExplicit() && !subject.metaIsDefinite())
						{
							blank_variables << subject.metaValue().toString();
						}
						if (object.metaIsExplicit() && !object.metaIsDefinite())
						{
							blank_variables << object.metaValue().toString();
						}
					}
					// create a new model only if we're the first model or
					// if the operation has insertion blanks in it
					if(blank_variables.size() || i == 0)
					{
						foreach (const QString& blank, blank_variables)
						{
							columns.addColumn(blank, RDFVariable());
						}
						update_modules << QSharedCreator<TrackerUpdateModule>::create(
							sharedFromThis<TrackerService>(),
							columns,
							tracker_update,
							true,
							strategy & BatchMode,
							strategy & SyncOnCommit);
						update_modules_raw << update_modules[i].data();
					} else
					{
						update_modules << QSharedPointer<TrackerUpdateModule>();
						update_modules_raw << 0;
					}
				}

				QList<LiveNodes> models;
				for (int i = 0; i < operation_count; ++i)
					if(TrackerUpdateModule *active = update_modules_raw[i])
					{
						// Each TrackerUpdateModule has its own model (where it will eventually insert
						// the data), but the data comes from the same source.  Hence, the modules need
						// to know about each other.
						active->setSisterModules(update_modules_raw, i);
						models << LiveNodes(newTripleModel(strategy & 0, update_modules[i],
														   TripleModulePtr(),
														   parent));
					} else
						models << LiveNodes();
				return models;
			}

			void TrackerService::addRemoveResourceClause
					(RDFStatementList *removed_statements, RDFVariable const &res)
			{
				removed_statements->push_back(RDFStatement(res, rdf::type::iri(), rdfs::Resource::iri()));
				QVector<RDFProperty> const dp = res.derivedProperties();
				for(QVector<RDFProperty>::const_iterator cdpi = dp.begin(), cdpiend = dp.end()
						; cdpi != cdpiend; ++cdpi)
					if(cdpi->strategy() & RemoveSubjectProperty)
					{
						RDFProperty prop = cdpi->deepCopy();
						RDFVariable var = res.optional().bindSubject(prop);
						removed_statements->push_back(RDFStatement(var, prop.predicate(), res));
					}
			}

			void TrackerService::setVerbosity(unsigned verbosity)
			{
				RDFServiceBasic::setVerbosity(verbosity);
				BackEnds::Local::service()->setVerbosity(verbosity);
			}

			void TrackerService::setConsoleVerbosity(unsigned console_verbosity)
			{
				RDFServiceBasic::setConsoleVerbosity(console_verbosity);
				BackEnds::Local::service()->setConsoleVerbosity(console_verbosity);
			}

			void TrackerService::setLogVerbosity(unsigned log_verbosity)
			{
				RDFServiceBasic::setLogVerbosity(log_verbosity);
				BackEnds::Local::service()->setLogVerbosity(log_verbosity);
			}

			void TrackerService::setLogDir(QDir const &path)
			{
				RDFServiceBasic::setLogDir(path);
				BackEnds::Local::service()->setLogDir(path);
			}

			ClassUpdateSignaler *ClassUpdateSignaler::get(QUrl class_uri)
			{
				return TrackerService::get()->dispatcherFromClass(class_uri.toString());
			}

			ClassUpdateSignaler *ClassUpdateSignaler::get
					(QString ontology_prefix, QString class_name)
			{
				return TrackerService::get()->dispatcherFromClass(TrackerService::get()
						->prefixSuffix(ontology_prefix, class_name).toString());
			}
		}
	}
}
