/*
 * This file is part of LibQtTracker project
 *
 * Copyright (C) 2009, Nokia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
/*
 * rdfservice_p.cpp
 *
 *  Created on: Mar 30, 2009
 *      Author: Iridian Kiiskinen <ext-iridian.kiiskinen at nokia.com>
 */

#include "rdfservice_p.h"
#include "rowstoremodel_p.h"
#include "include/sopranolive/liveautogenerationbase.h"

namespace SopranoLive
{

	RDFServicePtr RDFGraphBasic::service() const
	{
            if( external_parent_ref_ )
                return external_parent_ref_;

            //return parent_;
            // in this case parent is self so toStrongRef will always succee
            return parent_.toStrongRef();
	}

	QUrl RDFGraphBasic::graphIri() const
	{
		return graph_iri;
	}

	QVector<QStringList> RDFGraphBasic::rawExecuteQuery(QString const &query, int type)
	{
		return parent_.toStrongRef()->rawExecuteQueryOnGraph(query, type, *this);
	}
	void RDFGraphBasic::rawInsertTriple
			( QString const &subject, QString const &predicate, QString const &object)
	{
		return parent_.toStrongRef()->rawInsertTripleToGraph(subject, predicate, object, *this);
	}
	void RDFGraphBasic::rawDeleteTriple
			( QString const &subject, QString const &predicate, QString const &object)
	{
		return parent_.toStrongRef()->rawDeleteTripleFromGraph(subject, predicate, object, *this);
	}
	void RDFGraphBasic::rawLoad(QUrl const &file_uri)
	{
		return parent_.toStrongRef()->rawLoadOnGraph(file_uri, *this);
	}

	QUrl RDFGraphBasic::createUniqueIri(QString stem_text)
	{
		return parent_.toStrongRef()->createUniqueIri(stem_text);
	}

	LiveNode RDFGraphBasic::createLiveNode(CppType type)
	{
		return liveNode
			( type.metaTypeId() == qMetaTypeId<QUrl>() || type.metaTypeId() >= QMetaType::User
				? Node(createUniqueIri("generated_unique_id"))
				: rawNodeFromString(QString(), type)
			);
	}

	void RDFGraphBasic::addPrefix(QString prefix, QUrl value)
	{
		QString ns = value.toString();
		prefixes[prefix] = ns;
		reverse_prefixes[ns] = prefix;
	}

	QUrl RDFGraphBasic::prefixSuffix(QString prefix, QString suffix)
	{
		Prefixes::iterator pi = prefixes.find(prefix);
		if(pi != prefixes.end())
			return *pi + suffix;
		if(parent_)
                        parent_.toStrongRef()->prefixSuffix(prefix, suffix);
		qWarning() << "SopranoLive: unrecognized iri prefix \"" + prefix + "\"";
		return QUrl();
	}

	QPair<QString, QString> RDFGraphBasic::splitToPrefixSuffix(QString const &iri)
	{
		Prefixes::iterator pi = reverse_prefixes.upperBound(iri);
		if(pi != reverse_prefixes.begin())
		{
			--pi;
			if(iri.midRef(0, pi.key().size()) == pi.key())
				return qMakePair(pi.value(), iri.mid(pi.key().size()));
		}
		return qMakePair(QString(), iri);
	}

	LiveNode RDFGraphBasic::implLiveNode
			( LiveNode const &node, LiveNode const &type, bool strict
			, QList<QUrl> const &resource_contexts)
	{
		if(!node.isValid())
			return node;

		// if this graph implements the any of the supplied resource_contexts,
		// it is given that resource exists and has requested types, thus skip checks
		if(!node.isLiteral()
				&& (!resource_contexts.size() || !isAnyContextImplemented(resource_contexts)))
		{
			RDFVariable type_var(type.isValid()
					? type.variable()
					: RDFVariable(rdfs::Resource::iri()));
			if(strict)
			{
				RDFVariable checker = RDFVariable::fromType(type_var);
				checker == node;
				if(!modelVariable(checker)->rowCount())
					return LiveNode();
			} else
				// derived type implies all base types, suffices to add only it
				executeQuery(service_context_data_->update()
						.addInsertion(node, rdf::type::iri(), type_var));
		}

		if(node.isLive() && node->graph() == this)
			// checks have been made and node already is on this graph, return it directly
			return node;

		// otherwise, create a new node
		return newLiveResource(node)->initialSharedFromThis<LiveResource>();
	}

	LiveNodes RDFGraphBasic::implLiveNodes
	( LiveNodeList const &urls, RDFVariable const &object_info,
	  RDFStrategyFlags model_strategy, bool strict,
	  QSharedPointer<QObject> const &parent)
	{
		// TODO: something maybe
		Q_UNUSED(urls); Q_UNUSED(object_info); Q_UNUSED(model_strategy); Q_UNUSED(strict); Q_UNUSED(parent);
		critical(1) << "Here is no code; it's in RDFGraphCache";
		return LiveNodes();
	}

	ResourceOfService *RDFGraphBasic::newLiveResource(Node const &node)
	{
		return new ResourceOfService(sharedFromThis<RDFGraphBasic>(), node);
	}


	Node RDFGraphBasic::rawNodeFromString(QString string, CppType type)
	{
		return parent_.toStrongRef()->rawNodeFromString(string, type);
	}

	QString RDFGraphBasic::rawStringFromNode(Node const &node)
	{
		return parent_.toStrongRef()->rawStringFromNode(node);
	}

	void RDFGraphBasic::insertTriple
			( LiveNode const &subject, LiveNode const &predicate, LiveNode const &object)
	{
		return parent_.toStrongRef()->insertTripleToGraph(subject, predicate, object, *this);
	}
	void RDFGraphBasic::deleteTriple
			( LiveNode const &subject, LiveNode const &predicate, LiveNode const &object)
	{
		return parent_.toStrongRef()->deleteTripleFromGraph(subject, predicate, object, *this);
	}

	LiveNodes RDFGraphBasic::modelQuery
			( QString const &select, QStringList const &columns, RDFStrategyFlags model_strategy
			, QSharedPointer<QObject> const &parent, TripleModulePtr const &triple_module)
	{
		return parent_.toStrongRef()->modelQuery
				(select, columns, model_strategy, parent, triple_module);
	}
	LiveNodeModelBasePtr RDFGraphBasic::baseImplModelQuery
			( RDFDerivedSelect *select, RDFStrategyFlags model_strategy
			, QSharedPointer<QObject> const &parent, TripleModulePtr const &triple_module)
	{
		return parent_.toStrongRef()->baseImplModelQuery
				(select, model_strategy, parent, triple_module);
	}
	QList<LiveNodes> RDFGraphBasic::executeQuery
			( RDFQuery const &query, RDFStrategyFlags model_strategy
			, QSharedPointer<QObject> const &parent)
	{
		return parent_.toStrongRef()->executeQuery(query, model_strategy, parent);
	}

	LiveNodes RDFGraphBasic::implModelResources(RDFProperty &resource_info)
	{
		LiveNode type = liveNode(resource_info.predicate().metaValue());
		resource_info.setPredicate(rdf::type::iri());
		resource_info.setStrategy(resource_info.strategy() | RDFStrategy::Subject);
		return liveNode(type)->implGetProperties(resource_info);
	}

	void RDFGraphBasic::addRemoveResourceClause(RDFStatementList *removed_statements, RDFVariable const &res)
	{
		RDFStatement sts = RDFStatement(RDFVariable(), RDFVariable(), RDFVariableLink(res));
		sts.subject().property(sts.predicate()) = sts.object();
		RDFStatement sto = RDFStatement(RDFVariableLink(res));
		sto.subject().property(sto.predicate()) = sto.object();
		*removed_statements << sts << sto;
	}

	RDFStatementList *RDFGraphBasic::recurseRemoveResources
			(RDFStatementList *removed_statements, RDFVariable const &res
			, QLinkedList<RDFVariable> *post_removes)
	{
		QVector<RDFProperty> const dp = res.derivedProperties();
		for(QVector<RDFProperty>::const_iterator cdpi = dp.begin(), cdpiend = dp.end()
				; cdpi != cdpiend; ++cdpi)
		{
			RDFStrategyFlags str = cdpi->strategy();
			if(!(str & (RDFStrategy::Owns | RDFStrategy::Shares)))
				continue;

			RDFVariable derived =
					( (str & RDFStrategy::NonOptionalValued) || !(str & RDFStrategy::Owns)
							? RDFVariableLink(res) : RDFVariableLink(res).optional())
					.property(cdpi->deepCopy());
			if(str & RDFStrategy::Owns)
			{
				if(str & RDFStrategy::NonMultipleValued)
					recurseRemoveResources(removed_statements, derived, post_removes);
				else
				{
					RDFStatementList derived_removes;
					executeQuery(RDFUpdate(service_context_data_->update())
							.addDeletion(*recurseRemoveResources
									(&derived_removes, derived, post_removes)));
					;
				}
			} else
			if(str & RDFStrategy::Shares)
			{
				// Expensive, but can't really be avoided unless backend supports sophisticated
				// conditional cascading delete.
				// This is because either we need to create a horribly convoluted query for
				// checking whether a shared object being deleted is shared by a resource that is
				// _not_ getting deleted, or then we get the list of shared resources for which
				// a sharer was destroyed, and then requery if all they no longer have sharers
				// after that.
				LiveNodes actual_deriveds = modelVariables(RDFVariableList() << derived);
				if(actual_deriveds->rowCount())
				{
					RDFVariable var = RDFVariable::fromContainer(actual_deriveds);
					var.setDerivedProperties(derived.derivedProperties());
					// TODO: huh.. we should have a reverseProperty here
					var.optional().subject(cdpi->deepCopy()).doesntExist();
					post_removes->push_back(var);
				}
			}
		}
		addRemoveResourceClause(removed_statements, res);
		return removed_statements;
	}

	void RDFGraphBasic::removeResources(RDFVariable const &variable)
	{
		QLinkedList<RDFVariable> removes;
		removes << variable;
		RDFTransactionPtr tx = createTransaction();
		while(1)
		{
			RDFStatementList removed_statements;
			executeQuery
					(RDFUpdate(service_context_data_->update())
					.addDeletion(*recurseRemoveResources
							(&removed_statements, removes.front(), &removes)));
			removes.pop_front();
			tx->commit();
			if(!removes.size())
				return;
			if(!(tx = createTransaction(RDFTransaction::Exclusive)))
			{
				warning(1) << "RDFService::removeResources using recursive sharing removes requires that no transaction is active on the service";
				return;
			}
		}
	}

	TripleUpdateContext::Mode RDFGraphBasic::registerForTripleUpdates
			(SharedTripleUpdateContext const &context)
	{
		// TODO: context unused
		Q_UNUSED(context);
		return TripleUpdateContext::Inactive;
	}

	bool RDFGraphBasic::isAnyContextImplemented(QList<QUrl> const &contexts)
	{
		for(QList<QUrl>::const_iterator cci = contexts.begin(), cciend = contexts.end()
				; cci != cciend; ++cci)
			if(implemented_contexts_.find(cci->toString()) != implemented_contexts_.end())
				return true;
		return false;
	}

	void RDFGraphBasic::addImplementedContext(QUrl const &context)
	{
		++implemented_contexts_[context.toString()];
	}


	RDFServiceBasic::Transaction::Transaction
			( QSharedPointer<RDFServiceBasic> const &service, RDFTransaction::Mode mode)
		: TransactionBase(mode), service_(service), active_subs_(0)
	{
		service_->debug(4) << "Transaction TX(" << this << ") created";
	}

	RDFServiceBasic::Transaction::~Transaction()
	{
		if(isActive())
		{
			if(service_->verbosity() >= 2)
				service_->warning(2) << "~RDFTransaction: rolling back";
			if(mode_ & RDFTransaction::CommitOnDestroy)
				commit();
			else
				rollback();
		}
		service_->debug(4) << "Transaction TX(" << this << ") destroyed";
	}

	bool RDFServiceBasic::Transaction::commit(bool is_blocking)
	{
		if(!isActive())
		{
			service_->warning(2) << "RDFTransaction::commit: "
					"transaction is already inactive, cannot commit";
			return false;
		}

		// Nullify the service's transaction so that when we executeQuery, the
		// service doesn't try to append the query to its transaction but
		// really executes it.
		service_->transaction_ = 0;
		inactivate();

		if(update_.isEmpty())
		{
			Q_EMIT commitFinished();
			return true;
		}

		QList<LiveNodes> res = service_->executeQuery(update_, mode_ & BackEndFlagsMask);
		if(res.size())
		{
			pending_commit_ = res[0];
			res.clear();
			connect( pending_commit_.model(), SIGNAL(modelUpdated())
					, this, SIGNAL(commitFinished()));
			connect( pending_commit_.model(), SIGNAL(error
						( QString const &, RDFStrategyFlags
						, RDFStrategyFlags, QModelIndex const &))
					, this, SIGNAL(commitError(QString)));
			if(is_blocking)
				pending_commit_->refreshModel(LiveNodeModel::Block);
		}
		update_ = RDFUpdate();
		return true;
	}

	// TODO: implement signaling
	bool RDFServiceBasic::Transaction::rollback(bool is_blocking)
	{
		Q_UNUSED(is_blocking);
		if(!isActive())
		{
			service_->warning(2) << "RDFTransaction::rollback: "
					"transaction has already been committed/rolled back";
			return false;
		}
		service_->transaction_ = 0;
		inactivate();
		mode_ &= ~RDFTransaction::PendingRollback;

		update_ = RDFUpdate();

		emit rollbackFinished();
		return true;
	}

	bool RDFServiceBasic::Transaction::reinitiate()
	{
		if(!service_->transaction_)
		{
			service_->transaction_ = this;
			update_ = service_->service_context_data_->update();
			activate();
			return true;
		} else
			service_->warning(2) << "RDFTransaction::initiate: "
					"a transaction is already pending on target service";
		return false;
	}

	// TODO: implement blocking commits and signaling
	bool RDFServiceBasic::Transaction::commitAndReinitiate(bool is_blocking)
	{
		if(service_->transaction_ == this && !commit(is_blocking))
			return false;
		return reinitiate();
	}

	// TODO: implement signaling
	bool RDFServiceBasic::Transaction::rollbackAndReinitiate(bool is_blocking)
	{
		if(service_->transaction_ == this && !rollback(is_blocking))
			return false;
		return reinitiate();
	}

	//! check if the given \a child transaction can be made a child of this
	bool RDFServiceBasic::Transaction::canAdopt(RDFTransaction::Mode const &child_mode) const
	{
		if((mode_ | child_mode) & RDFTransaction::Exclusive)
			return false;
		return (mode_ & BackEndFlagsMask) == (child_mode & BackEndFlagsMask);
	}

	RDFServicePtr RDFServiceBasic::Transaction::service() const
	{
		return service_;
	}

	RDFServiceBasic::InPlaceSubTransaction::InPlaceSubTransaction
			(QSharedPointer<Transaction> const &parent, RDFTransaction::Mode mode)
		: TransactionBase(mode), parent_()
	{
		setParent(parent);
		parent_->service_->debug(4) << "InPlaceSubTransaction IPSTX(" << this << ") created";
	}

	RDFServiceBasic::InPlaceSubTransaction::~InPlaceSubTransaction()
	{
		parent_->service_->debug(4) << "InPlaceSubTransaction IPSTX(" << this << ") destroyed,"
				<< ((mode_ & RDFTransaction::CommitOnDestroy) ? " committing" : " rolling back");
		if(mode_ & RDFTransaction::CommitOnDestroy)
			commit();
		else
			rollback();
	}

	RDFServicePtr RDFServiceBasic::InPlaceSubTransaction::service() const
	{ return parent_->service(); }

	RDFTransaction::Mode RDFServiceBasic::InPlaceSubTransaction::mode() const
	{
		return mode_ |
			(parent_->active_subs_ - (isActive() ? 1 : 0) ? RDFTransaction::OtherActive : 0);
	}


	void RDFServiceBasic::InPlaceSubTransaction::setParent
			(QSharedPointer<Transaction> const &parent)
	{
		if(parent_)
			parent_->disconnect(this);
		parent_ = parent;
		Transaction *p = parent.data();
		connect(p, SIGNAL(commitFinished()), this, SIGNAL(commitFinished()));
		connect(p, SIGNAL(rollbackFinished()), this, SIGNAL(rollbackFinished()));
		connect(p, SIGNAL(commitError(QString)), this, SIGNAL(commitError(QString)));
	}

	bool RDFServiceBasic::InPlaceSubTransaction::isInPlacePendingRollback() const
	{ return parent_->isInPlacePendingRollback(); }

	bool RDFServiceBasic::InPlaceSubTransaction::commit(bool is_blocking)
	{
		if(!isActive())
			return false;
		inactivate();
		if(--parent_->active_subs_)
			return false;
		if(parent_->isPendingRollback())
		{
			parent_->rollback(is_blocking);
			return false;
		}
		return parent_->commit(is_blocking);
	}
	bool RDFServiceBasic::InPlaceSubTransaction::rollback(bool is_blocking)
	{
		if(!isActive())
			return false;
		inactivate();
		if(--parent_->active_subs_)
		{
			parent_->mode_ |= RDFTransaction::PendingRollback;
			return false;
		}
		return parent_->rollback(is_blocking);
	}
	bool RDFServiceBasic::InPlaceSubTransaction::reinitiate()
	{
		if(isActive())
			return false;
		if(!parent_->active_subs_)
		{
			Transaction *active = parent_->service_->transaction_;
			if(!active)
				parent_->reinitiate();
			else if(!active->canAdopt(mode_))
				return false;
			else
				setParent(parent_->service_->transaction_->sharedFromThis<Transaction>());
		}
		parent_->active_subs_++;
		activate();
		return true;
	}

	// TODO: think these through. Currently, a simple solution.
	bool RDFServiceBasic::InPlaceSubTransaction::commitAndReinitiate(bool is_blocking)
	{ return commit(is_blocking) && reinitiate(); }
	// TODO: think these through. Currently, a simple solution.
	bool RDFServiceBasic::InPlaceSubTransaction::rollbackAndReinitiate(bool is_blocking)
	{ return rollback(is_blocking) && reinitiate(); }

	RDFServiceBasic::RDFServiceBasic
			( QString name, QList<QUrl> const &implemented_contexts
			, QSharedPointer<RDFServiceBasic> const &parent
			, ServiceContextDataPtr const &ctxt)
		: RDFGraphBasic(parent, ctxt)
		, name_(name), console_verbosity_(1), log_verbosity_(0)
		, log_path_()
		, log_name_(name_ + "_" + QDateTime::currentDateTime().toString(Qt::ISODate) + ".log")
		, log_file_()
		, transaction_()
	{
            /*if(!this->parent_)
                        this->parent_ = initialSharedFromThis<RDFServiceBasic>();
                        */
		addImplementedContext(Contexts::NoChecks::iri());
		for(QList<QUrl>::const_iterator cici = implemented_contexts.begin()
				, ciciend = implemented_contexts.end(); cici != ciciend; ++cici)
			addImplementedContext(*cici);
		loadPrefixes();
		attributes_["debug_message_count"] = QVariant(0);
	}

	RDFServiceBasic::~RDFServiceBasic()
	{
		if(verbosity() >= 3)
			outputDebug("RDFServiceBasic(" + name_ + ") destroyed", 3);
	}


	void RDFServiceBasic::loadPrefixes()
	{
		// TODO: Think, is this sane? Are more needed? Less?
		// answer to above: prefix system might be unnecessary,
		// conveniece namespaces should be used instead
		// answer to above: prefix system is internally required in
		// converting full uris back to prefixes
		// conclusion: more is better, cannot hurt and the ontologies are stable
		addPrefix("rdf", QUrl("http://www.w3.org/1999/02/22-rdf-syntax-ns#"));
		addPrefix("rdfs", QUrl("http://www.w3.org/2000/01/rdf-schema#"));

		addPrefix("a", QUrl("http://www.w3.org/2000/10/annotation-ns#"));

		addPrefix("xsd", QUrl("http://www.w3.org/2001/XMLSchema#"));

		addPrefix("owl", QUrl("http://www.w3.org/2002/07/owl#"));

		addPrefix("skos", QUrl("http://www.w3.org/2004/02/skos/core#"));

		addPrefix("geo", QUrl("http://www.w3.org/2003/01/geo/wgs84_pos#"));

		addPrefix("dc", QUrl("http://purl.org/dc/elements/1.1/"));
		addPrefix("dcterms", QUrl("http://purl.org/dc/terms/"));
		addPrefix("dctype", QUrl("http://purl.org/dc/dcmitype/"));
		addPrefix("dcam", QUrl("http://purl.org/dc/dcam/"));

		addPrefix("foaf", QUrl("xmlns.com/foaf/0.1/"));
	}

	void RDFServiceBasic::insertTripleToGraph
			( LiveNode const &subject, LiveNode const &predicate, LiveNode const &object
			, RDFGraphImplBase const &/*graph*/)
	{
		if(!transaction_)
		{
			RDFUpdate insertion = service_context_data_->update();
			insertion.addInsertion(subject, predicate, object);
			executeQuery(insertion);
		} else
			transaction_->update_.addInsertion(subject, predicate, object);
		// TODO: implement graphs
		//rawInsertTripleToGraph
		//		( stringFromNode(subject), stringFromNode(predicate), stringFromNode(object)
		//		, graph);
	}

	void RDFServiceBasic::deleteTripleFromGraph
			( LiveNode const &subject, LiveNode const &predicate, LiveNode const &object
			, RDFGraphImplBase const &/*graph*/)
	{
		if(!transaction_)
		{
			RDFUpdate deletion = service_context_data_->update();
			deletion.addDeletion(subject, predicate, object);
			executeQuery(deletion);
		} else
			transaction_->update_.addDeletion(subject, predicate, object);
		// TODO: implement graphs
		//rawDeleteTripleFromGraph
		//		( stringFromNode(subject), stringFromNode(predicate), stringFromNode(object)
		//		, graph);
	}

	RDFTransactionPtr RDFServiceBasic::createTransaction(RDFTransaction::Mode mode)
	{
		QSharedPointer<Transaction> mid;
		bool initiate = !(mode & RDFTransaction::Inactive);
		mode |= RDFTransaction::Inactive;

		if(transaction_ && transaction_->canAdopt(mode))
			mid = transaction_->sharedFromThis<Transaction>();
		else if(!initiate || !transaction_)
			mid = QSharedCreator<Transaction>::create(sharedFromThis<RDFServiceBasic>(), mode);
		else
			return RDFTransactionPtr();

		RDFTransactionPtr ret = mode & RDFTransaction::Exclusive
			? RDFTransactionPtr(mid)
			: RDFTransactionPtr(QSharedCreator<InPlaceSubTransaction>::create
					(mid, mode));

		if(initiate)
			ret->reinitiate();
		return ret;
	}

	RDFTransactionPtr RDFServiceBasic::pendingTransaction() const
	{
		if(transaction_)
			return transaction_->sharedFromThis<RDFTransaction>();
		return RDFTransactionPtr();
	}

	RDFServicePtr RDFServiceBasic::createChildService()
	{
		// TODO: implement subservices
		return RDFServicePtr();
	}

	unsigned RDFServiceBasic::verbosity() const
	{
		return std::max(console_verbosity_, log_verbosity_);
	}

	void RDFServiceBasic::setVerbosity(unsigned verbosity)
	{
		setConsoleVerbosity(verbosity);
	}

	unsigned RDFServiceBasic::consoleVerbosity() const
	{
		return console_verbosity_;
	}

	void RDFServiceBasic::setConsoleVerbosity(unsigned console_verbosity)
	{
		console_verbosity_ = console_verbosity;
	}

	unsigned RDFServiceBasic::logVerbosity() const
	{
		return log_verbosity_;
	}

	void RDFServiceBasic::setLogVerbosity(unsigned new_log_verbosity)
	{
		if(!log_verbosity_ && new_log_verbosity)
		{
			setLogDir(log_path_);
			log_file_.setFileName(log_path_.filePath(log_name_));
			if(!log_file_.open(QIODevice::ReadWrite | QIODevice::Text | QIODevice::Append))
				warning() << "RDFServiceBasic::setLogVerbosity: "
						"unable to open log file for writing("
						<< log_path_.filePath(log_name_) << ")";
		} else if(log_verbosity_ && !new_log_verbosity)
		{
			log_file_.close();
		}
		log_verbosity_ = new_log_verbosity;
	}

	void RDFServiceBasic::setLogDir(QDir const &path)
	{
		log_path_ = path;
		if(int current_verb = log_verbosity_)
		{
			setLogVerbosity(0);
			setLogVerbosity(current_verb);
		}
	}


	bool RDFServiceBasic::outputWarning(QString const &message, unsigned verbosity) const
	{
		if(!message.size())
			return true;
		if(this->console_verbosity_ < verbosity && this->log_verbosity_ < verbosity)
			return false;
		if(this->console_verbosity_ >= verbosity)
			qWarning() << (name_ + "(" + QString::number(verbosity) + "): ").toAscii().data()
					<< message.toAscii().data();
		if(this->log_verbosity_ >= verbosity)
			QTextStream(&log_file_)
					<< ("qWarning(" + QString::number(verbosity) + "): ").toAscii().data()
					<< message.toAscii().data() << "\n";
		return true;
	}

	QTime initDebugCounter()
	{
		QTime ret;
		ret.start();
		return ret;
	}
	static QTime debug_counter_ = initDebugCounter();

	bool RDFServiceBasic::outputDebug(QString const &message, unsigned verbosity) const
	{
		if(!message.size())
			return true;
		if(this->console_verbosity_ < verbosity && this->log_verbosity_ < verbosity)
			return false;
		if(this->console_verbosity_ >= verbosity)
		{
			int msec = debug_counter_.elapsed();
			int sec = msec / 1000;
			msec -= sec * 1000;
			int dmsgc = attributes_["debug_message_count"].toInt();
			attributes_["debug_message_count"] = QVariant(dmsgc + 1);
			qDebug().nospace()
					<< (name_ + "(" + QString::number(verbosity) + ")[").toAscii().data()
					<< right
					<< qSetFieldWidth(5) << sec
					<< qSetFieldWidth(0) << "."
					<< qSetFieldWidth(3) << qSetPadChar('0') << msec
					<< qSetFieldWidth(0) << qSetPadChar(' ')
					<< "/"
					<< qSetFieldWidth(5) << dmsgc
					<< qSetFieldWidth(0)
					<< "]: " << message.toAscii().data();
		}
		if(this->log_verbosity_ >= verbosity)
			QTextStream(&log_file_)
					<< ("qDebug(" + QString::number(verbosity) + "): ").toAscii().data()
					<< message.toAscii().data() << "\n";
		return true;
	}
	bool RDFServiceBasic::outputCritical(QString const &message, unsigned verbosity) const
	{
		if(!message.size())
			return true;
		if(this->console_verbosity_ < verbosity && this->log_verbosity_ < verbosity)
			return false;
		if(this->console_verbosity_ >= verbosity)
			qCritical() << (name_ + "(" + QString::number(verbosity) + "): ").toAscii().data()
					<< message.toAscii().data();
		if(this->log_verbosity_ >= verbosity)
			QTextStream(&log_file_)
					<< ("qCritical(" + QString::number(verbosity) + "): ").toAscii().data()
					<< message.toAscii().data() << "\n";
		return true;
	}

	QVariant RDFServiceBasic::serviceAttribute(QString const &attribute_name) const
	{
		AttributeMap::const_iterator ai = attributes_.find(attribute_name);
		if(ai != attributes_.end())
			return *ai;
		return QVariant();
	}

	QVariantMap RDFServiceBasic::serviceAttributes() const
	{
		return attributes_;
	}

	//! Accepts all attributes which make it here.
	bool RDFServiceBasic::setServiceAttribute
			( QString const &service_attribute, QVariant const &value)
	{
		if(!value.isValid())
			attributes_.remove(service_attribute);
		attributes_[service_attribute] = value;
		return true;
	}

	// TODO: implement shared parent aggregation
	RowStoreModelPtr RDFServiceBasic::newRowStoreModel
			( RDFStrategyFlags strategy_flags, ContentModulePtr const &content_module
			, QSharedPointer<QObject> const &parent)
	{
		Q_UNUSED(parent);
		RowStoreModelPtr ret = QSharedCreator<RowStoreModel>::create
				(sharedFromThis<RDFServiceBasic>(), strategy_flags);
		ret->changeContentModule(content_module);
		return ret;
	}

	// TODO: implement shared parent aggregation
	TripleModelPtr RDFServiceBasic::newTripleModel
			( RDFStrategyFlags strategy_flags, ContentModulePtr const &content_module
			, TripleModulePtr const &cache_module, QSharedPointer<QObject> const &parent)
	{
		Q_UNUSED(parent);
		TripleModelPtr ret = QSharedCreator<TripleModel>::create
				(sharedFromThis<RDFServiceBasic>(), strategy_flags);
		ret->changeContentModule(content_module);
		ret->changeTripleModule(cache_module);
		return ret;
	}

}
